(*
 * The following is a substitute to the Jane Street Core library used in earlier versions of RegSTAB.
 * It matches quite precisely the interface of Jane Street Core.
 *
 * A few pieces of code are even copy-pasted from Core whose licence is GNU
 * Lesser General Public License version 2.1 and whose copyright is as follows:
 *
 * Copyright (C) 2008-
 * 
 *   Jane Street Holding, LLC
 *   1 New York Plaza, 33rd Floor 
 *   New York, NY 10004
 *   USA
 * 
 *   email: opensource@janestcapital.com
 * 
 * One can find Jane Street Core library can be found at
 * <http://ocaml.janestreet.com/?q=node/13>.
 *)

external ident : 'a -> 'a = "%identity"
let (<||) f g = fun x -> f (g x)

module Hashtbl =
  struct
    module H = MoreLabels.Hashtbl
    include H
    let find_exn t key = H.find t key
    let find t key = try let res = H.find t key in Some res with Not_found -> None
  end

module Tuple2 =
  struct
    type ('a,'b) t = 'a * 'b
    let get1 (x,_) = x
    let get2 (_,y) = y
  end

module Tuple3 =
  struct
    type ('a,'b,'c) t = 'a * 'b * 'c
    let get1 (x,_,_) = x
    let get2 (_,y,_) = y
    let get3 (_,_,z) = z
  end

module List =
  struct
    module L = StdLabels.List
    include L
    let fold t ~init ~f = fold_left t ~f ~init
    let map l ~f = L.rev (L.rev_map ~f l)
    let dedup list =
      let sorted = L.sort ~cmp:(fun x y -> compare y x) list in
      let rec loop list accum = match list with
        |[]           -> accum
        |hd::[]       -> loop [] (hd::accum)
        |hd1::hd2::tl -> if hd1 = hd2 then loop (hd2::tl) accum else loop (hd2::tl) (hd1::accum)
      in
      loop sorted []
    let range ?(stride=1) low high =
      if stride <= 0 then invalid_arg "Softcore.List.range: stride must be positive";
      let rec loop low high accum = if low >= high then accum else loop (low + stride) high (low::accum) in
      L.rev (loop low high [])
    let reduce l ~f = match l with
      |[]       -> raise (Invalid_argument "List.reduce")
      |init::l' -> L.fold_left ~init ~f l'
    let append l1 l2 = L.rev_append (L.rev l1) l2
    let rev_filter_map l ~f =
      let rec loop l acc = match l with
        |[]     -> acc
        |hd::tl -> loop tl (match f hd with Some x -> x::acc | None -> acc)
      in
      loop l []
    let filter_map l ~f = List.rev (rev_filter_map l ~f)
  end

module Option =
  struct
    let value t ~default = match t with None -> default | Some x -> x
    let is_some = function Some _ -> true  | _ -> false
    let is_none = function Some _ -> false | _ -> true
  end

module Caml =
  struct
    module Map = Map
    module Set = Set
    module MoreLabels = MoreLabels
    module StdLabels = StdLabels
  end

module Exn =
  struct
    type t = exn
    exception Finally of t * t
    let protectx ~f x ~(finally : 'a -> unit) =
      let res =
        try f x
        with exn ->
          (try finally x with final_exn -> raise (Finally (exn, final_exn)));
          raise exn
      in
      finally x;
      res
  end

module Map =
  struct
    type ('key,'a) t = Empty | Node of ('key,'a) t * 'key * 'a * ('key,'a) t * int

    let height = function Empty -> 0 | Node(_,_,_,_,h) -> h

    let create l x d r =
      let hl = height l and hr = height r in
      Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

    let singleton key data = create Empty key data Empty

    let bal l x d r =
      let hl = match l with Empty -> 0 | Node(_,_,_,_,h) -> h in
      let hr = match r with Empty -> 0 | Node(_,_,_,_,h) -> h in
      if hl > hr + 2 then begin
        match l with
        |Empty               -> invalid_arg "Map.bal"
        |Node(ll,lv,ld,lr,_) -> 
            if height ll >= height lr then create ll lv ld (create lr x d r)
            else begin match lr with
              |Empty                   -> invalid_arg "Map.bal"
              |Node(lrl,lrv,lrd,lrr,_) -> create (create ll lv ld lrl) lrv lrd (create lrr x d r)
            end
      end else if hr > hl + 2 then begin
        match r with
        |Empty -> invalid_arg "Map.bal"
        |Node(rl, rv, rd, rr, _) ->
            if height rr >= height rl then create (create l x d rl) rv rd rr
            else begin match rl with
              |Empty                       -> invalid_arg "Map.bal"
              |Node(rll, rlv, rld, rlr, _) -> create (create l x d rll) rlv rld (create rlr rv rd rr)
            end
      end else Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

    let empty = Empty

    let is_empty = function Empty -> true | _ -> false

    let rec add ~key:x ~data = function
      |Empty           -> Node(Empty, x, data, Empty, 1)
      |Node(l,v,d,r,h) -> 
          let c = Pervasives.compare x v in
          if c = 0 then Node(l, x, data, r, h)
          else if c < 0 then bal (add ~key:x ~data l) v d r
          else bal l v d (add ~key:x ~data r)

    let rec find t x =
      match t with
      |Empty           -> None
      |Node(l,v,d,r,_) -> 
          let c = Pervasives.compare x v in
          if c = 0 then Some d
          else find (if c < 0 then l else r) x

    let rec find_exn t x = match find t x with None -> raise Not_found | Some data -> data

    let rec mem t x =
      match t with
      |Empty           -> false
      |Node(l,v,_,r,_) -> 
          let c = Pervasives.compare x v in
          c = 0 || mem (if c < 0 then l else r) x

    let rec min_elt = function
      |Empty                -> None
      |Node (Empty,k,d,_,_) -> Some (k, d)
      |Node (l,_,_,_,_)     -> min_elt l

    let rec min_elt_exn t = match min_elt t with None -> raise Not_found | Some v -> v

    let rec cardinal = function Empty -> 0 | Node (l, _, _, r, _) -> cardinal l + cardinal r + 1

    let rec rev_fold ~f t ~init:accu = match t with
      |Empty               -> accu
      |Node(l, v, d, r, _) -> rev_fold ~f l ~init:(f ~key:v ~data:d (rev_fold ~f r ~init:accu))

    let keys t = rev_fold ~f:(fun ~key ~data:_ list -> key::list) t ~init:[]

    let merge ~f t1 t2 =
      let all_keys = List.dedup (List.append (keys t1) (keys t2)) in
      List.fold_left ~init:empty all_keys
        ~f:(fun t key ->
              match f ~key (find t1 key) (find t2 key) with
              |None      -> t
              |Some data -> add ~key ~data t)
  end

module Set =
  struct
    type 'a t = ('a,unit) Map.t
    let empty = Map.empty
    let add t x = Map.add t ~key:x ~data:()
    let min_elt_exn t = let (x,_) = Map.min_elt_exn t in x
    let min_elt t = try Some (min_elt_exn t) with Not_found -> None
    let choose = min_elt
    let choose_exn = min_elt_exn
    let elements = Map.keys
    let cardinal = Map.cardinal
    let union t1 t2 = Map.merge t1 t2 ~f:(fun ~key data1 data2 -> match data1,data2 with Some(),_ | _,Some() -> Some () | _ -> None)
    let singleton x = Map.singleton x ()
  end

module String = 
  struct
    module S = StdLabels.String
    include S

    let split str ~on =
      let len = length str in
      let rec loop acc last_pos pos =
        if pos = -1 then
          sub str ~pos:0 ~len:last_pos :: acc
        else
          if str.[pos] = on then
            let pos1 = pos + 1 in
            let sub_str = sub str ~pos:pos1 ~len:(last_pos - pos1) in
            loop (sub_str :: acc) pos (pos - 1)
          else loop acc last_pos (pos - 1)
      in
      loop [] len (len - 1)
  end

let protectx = Exn.protectx

let read_wrap ?(binary = false) ~f fname =
  let ic =
    if binary then
      open_in_bin fname
    else
      open_in fname
  in
  protectx ic ~f ~finally:close_in

module Array = StdLabels.Array

let list_to_string ~to_string ?delim:(delim1,delim2=("[","]")) ?(sep=" ; ") xs = 
  let rec list_to_string' = function
    |[] -> ""
    |[x] -> to_string x
    |x::xs -> (to_string x) ^ sep ^ (list_to_string' xs)
  in
  delim1 ^ (list_to_string' xs) ^ delim2
